<?php
/**
 * Created by PhpStorm.
 * User: dennis
 * Date: 5/7/15
 * Time: 12:38 PM
 */

namespace Zitcom\IPUtils;


class Util {

    protected function __construct() {
        throw new \Exception('Static methods only');
    }

    /**
     * Calculates the CIDR notations from a given IPv4 range
     *
     * @param $start
     * @param $end
     * @return array
     */
    public static function cidrFromRange($start, $end) {
        $result = array();

        $nMin = ip2long($start);
        $nMax = ip2long($end);

        $bitsMin= 1;
        $maskMin= 1;
        $bitsMax= 1;
        $maskMax= 1;

        $i = 0;

        while (true) {
            $i++;
            while (0 == ($nMin & $maskMin)) {
                ($maskMin <<= 1);
                $maskMin++;
                $bitsMin++;
            }

            if ($nMax < ($nMin | $maskMin)) {
                break;
            }

            $result[] = long2ip($nMin) . '/' . (33 - $bitsMin);

            $nMin = $nMin | $maskMin;
            $nMin++;
        }

        $backwards = array();
        while (true) {
            while ($maskMax == ($nMax & $maskMax)) {
                $maskMax <<= 1;
                $maskMax++;
                $bitsMax++;
            }

            $nMax = $nMax & ~$maskMax;

            if ($nMax < $nMin) {
                break;
            }

            $backwards[] = long2ip($nMax--) . '/' . (33 - $bitsMax);

        }

        return array_merge($result, array_reverse($backwards));
    }

    public static function IpListFromNetmask ($network, $netmask) {
        $start = ip2long($network) & ip2long($netmask);
        $end = $start | (~ip2long($netmask) & ip2long('255.255.255.255'));

        $result = array();

        for ($ip = $start; $ip <= $end; $ip++) {
            $result[] = long2ip($ip);
        }

        return $result;
    }

    /**
     * @param $start
     * @param $end
     * @return string[]
     */
    public static function IpListFromRange($start, $end) {
        $start = ip2long($start);
        $end = ip2long($end);

        $result = array();

        for ($ip = $start; $ip <= $end; $ip++) {
            $result[] = long2ip($ip);
        }

        return $result;
    }

    public static function IpListFromCidr($cidr) {
        list($ip, $bits) = explode('/', $cidr);

        $netmask = ip2long('255.255.255.255') << (32 -$bits);


        $start = ip2long($ip) & $netmask;
        $end = $start | (~$netmask) & ip2long('255.255.255.255');

        $result = array();

        for ($ip = $start; $ip <= $end; $ip++) {
            $result[] = long2ip($ip);
        }


        return $result;
    }

    /**
     * @param $inet
     * @return string
     */
    public static function inet_to_bits($inet) {
        static $lookupTable = array(
            '0' => '0000',
            '1' => '0001',
            '2' => '0010',
            '3' => '0011',
            '4' => '0100',
            '5' => '0101',
            '6' => '0110',
            '7' => '0111',
            '8' => '1000',
            '9' => '1001',
            'a' => '1010',
            'b' => '1011',
            'c' => '1100',
            'd' => '1101',
            'e' => '1110',
            'f' => '1111',
        );

        $unpacked = unpack('H*', $inet);
        $unpacked = str_pad($unpacked[1], 32, '0', STR_PAD_LEFT);

        $binaryIp = '';

        foreach (str_split($unpacked) as $char) {
            $binaryIp .= $lookupTable[$char];
        }

        return $binaryIp;
    }

    /**
     * Given number of bits, it returns an IPv4 inverse netmask
     *
     * @param integer $bits
     * @return string
     */
    public static function inverseMask($bits) {
        return long2ip(abs(-1 << (32 - $bits))-1);
    }

    /**
     * Given number of bits, it returns an IPv4 mask
     *
     * @param $bits
     * @return string
     */
    public static function mask($bits) {
        return long2ip(-1 << (32 - (int)$bits));
    }

    public static function maskFromCidr($cidr) {
        list($ip, $bits) = explode('/', $cidr);

        return self::mask($bits);
    }

    /**
     * Supports both IPv4 and IPv6
     *
     * @param $checkIp
     * @param $cidr
     * @throws
     * @return bool
     */
    public static function IpInCidr($checkIp, $cidr) {
        if (false === strpos($cidr, '/')) {
            if (filter_var($cidr, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6)) {
                $cidr = $cidr . '/128';
            } else {
                $cidr = $cidr . '/32';
            }
        }

        list($ip, $bits) = explode('/', $cidr, 2);

        if (false === filter_var($ip, FILTER_VALIDATE_IP)) {
            return false;
        }

        if (filter_var($ip, FILTER_VALIDATE_IP, FILTER_FLAG_IPV4)) {
            $ipBits = 32;
        } else if (filter_var($ip, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6)) {
            $ipBits = 128;
        } else {
            throw new \Exception('Unsupported IP version');
        }

        $mask = gmp_init(str_repeat('1', $ipBits), 2);

        $mask = gmp_mul($mask, gmp_pow(2, ($ipBits - $bits)));

        $ip = gmp_import(inet_pton($ip));
        $checkIp = gmp_import(inet_pton($checkIp));

        if (gmp_export(gmp_and($ip, $mask)) === gmp_export(gmp_and($checkIp, $mask))) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * @param $bits
     * @param $workingNumber
     * @param $ipBits
     * @param $reserved
     * @return string[]
     */
    protected static function findSubnets($bits, &$workingNumber, $ipBits, $reserved = []) {
        $newBits = $bits;
        $newBits++;

        $result = [];

        if($newBits > $ipBits) {
            return $result;
        }

        for ($j = 0; $j < ($newBits - $bits) * 2; $j++) {
            if (static::ipInCidrMultiples([inet_ntop(gmp_export($workingNumber)) . '/' . $newBits], $reserved)) {
                $workingNumber = gmp_add($workingNumber, gmp_pow(2, $ipBits - $newBits));
                //break;
            } else if (static::ipInCidrMultiples($reserved, [inet_ntop(gmp_export($workingNumber)) . '/' . $newBits])) {
                // recursive
                $result = array_merge($result, static::findSubnets($newBits, $workingNumber, $ipBits, $reserved));
            } else {
                $result[] = inet_ntop(gmp_export($workingNumber)) . '/' . $newBits;
                $workingNumber = gmp_add($workingNumber, gmp_pow(2, $ipBits - $newBits));
            }
        }

        return $result;
    }

    /**
     * Checks if the $checkCidr is within the $cidr
     *
     * @param $checkCidr
     * @param $cidr
     * @return bool
     */
    public static function cidrInCidr($checkCidr, $cidr) {
        list($checkIp, $checkBits) = explode('/', $checkCidr);
        list($ip, $bits) = explode('/', $cidr);

        if ($checkBits < $bits) {
            return false;
        }

        return static::IpInCidr($checkIp, $cidr);
    }

    /**
     * Returns true if one of the checkips is in the cidrs
     *
     * @param array $checkCidrs
     * @param array $cidrs
     * @return bool
     */
    protected static function ipInCidrMultiples($checkCidrs = [], $cidrs = []) {
        foreach ($checkCidrs as $checkCidr) {
            list($checkIp, $checkBits) = explode('/', $checkCidr);

            foreach ($cidrs as $cidr) {
                list($ip, $bits) = explode('/', $cidr);

                if ($checkBits < $bits) {
                    continue;
                }

                if (static::IpInCidr($checkIp, $cidr)) {
                    return true;
                }
            }
        }

        return false;
    }

    /**
     * @param string $cidr
     * @param string[] $usedCidrs
     * @return string[]
     * @throws \Exception
     */
    public static function getAvailableNets($cidr, array $usedCidrs) {
        list($ip, $bits) = explode('/', $cidr);

        if (filter_var($ip, FILTER_VALIDATE_IP, FILTER_FLAG_IPV4)) {
            $ipBits = 32;
        } else if (filter_var($ip, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6)) {
            $ipBits = 128;
        } else {
            throw new \Exception('Invalid input ip');
        }

        $binaryIp = inet_pton($ip);
        $gmpNumber = gmp_import($binaryIp);

        $workingNumber = clone $gmpNumber;

        return static::findSubnets($bits, $workingNumber, $ipBits, $usedCidrs);
    }

    /**
     * @param $cidr
     * @return string
     */
    public static function gatewayFromCidr($cidr) {
        list($ip, $bits) = explode('/', $cidr);

        $netmask = ip2long('255.255.255.255') << (32 -$bits);


        return long2ip((ip2long($ip) & $netmask) + 1);
    }

    /**
     * @param $cidr
     * @return string
     */
    public static function networkFromCidr($cidr) {
        list($ip, $bits) = explode('/', $cidr);

        if (filter_var($ip, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6)) {
            $maxBits = 128;
        } else {
            $maxBits = 32;
        }

        $mask = gmp_init(str_repeat('1', $maxBits), 2);
        $mask = gmp_mul($mask, gmp_pow(2, ($maxBits - $bits)));

        $ip = gmp_import(inet_pton($ip));

        $network = gmp_and($ip, $mask);

        return inet_ntop(gmp_export($network));
    }

    /**
     * @param $cidr
     * @return string
     */
    public static function broadcastFromCidr($cidr) {
        list($ip, $bits) = explode('/', $cidr);

        $netmask = ip2long('255.255.255.255') << (32 -$bits);


        $start = ip2long($ip) & $netmask;
        return long2ip($start | (~$netmask) & ip2long('255.255.255.255'));
    }

    public static function bitsFromNetmask($netmask) {
        $long = ip2long($netmask);
        $base = ip2long('255.255.255.255');
        return 32-log(($long ^ $base)+1,2);
    }
}
